# flake8: NOQA
# "flake8: NOQA" to suppress warning "H104  File contains nothing but comments"

# TODO(okuta): Implement fliplr


# TODO(okuta): Implement flipud


# TODO(okuta): Implement roll


# TODO(okuta): Implement rot90
